"""Contains the ansXpl class"""
import weakref
import json

import numpy as np
from ansys.grpc.mapdl import ansys_kernel_pb2 as anskernel

from ansys.mapdl.core.errors import MapdlRuntimeError


class ansXpl():
    """ANSYS database explorer class.

    Examples
    --------
    >>> from ansys.mapdl.core import launch_mapdl
    >>> mapdl = launch_mapdl()
    >>> xpl = mapdl.xpl
    """

    def __init__(self, mapdl):
        from ansys.mapdl.core.mapdl_grpc import MapdlGrpc
        if not isinstance(mapdl, MapdlGrpc):  # pragma: no cover
            raise TypeError('Must be initialized using MapdlGrpc class')

        self._mapdl_weakref = weakref.ref(mapdl)
        self._filename = None
        self._open = False

    @property
    def _mapdl(self):
        """Return the weakly referenced instance of mapdl"""
        return self._mapdl_weakref()

    def open(self, filename, option=""):
        """Open an MAPDL file to explore

        Parameters
        ----------
        filename : str
            Name of the file to open.

        Returns
        -------
        mapdl_response : str
            Response from MAPDL.

        Examples
        --------
        >>> xpl.open('file.full')
        """
        self._filename = filename
        out = self._mapdl.run("*XPL,OPEN,%s,,%s" % (filename, option))
        self._open = True
        return out

    def close(self):
        """Close the MAPDL file after opening.

        Returns
        -------
        mapdl_response : str
            Response from MAPDL.

        Examples
        --------
        >>> xpl.close()
        """
        response = self._mapdl.run("*XPL,CLOSE")
        self._check_ignored(response)
        self._open = False
        return response

    def list(self, nlev=1):
        """List the records at the current level.

        Parameters
        ----------
        nlev: int
            Number of levels to recursively explore.

        Returns
        -------
        mapdl_response : str
            Response from MAPDL.

        Examples
        --------
        >>> xpl.list(1)
        """
        response = self._mapdl.run("*XPL,LIST,%d" % nlev)
        self._check_ignored(response)
        return response

    def _check_ignored(self, response):
        """Check for ignored in response"""
        if 'ignored' in response:
            raise MapdlRuntimeError(response)

    def help(self):
        """XPL help message

        Examples
        --------
        >>> print(xpl.help())
        """
        return self._mapdl.run("*XPL,HELP")

    def step(self, where):
        """Go down in the tree of records

        Parameters
        ----------
        where : str
            Path to follow. This path can be composed of several
            levels, for example ``"BRANCH1::SUBBRANCH2::.."``

        Returns
        -------
        mapdl_response : str
            Response from MAPDL.

        Examples
        --------
        >>> xpl.step('MASS')
        >>> print(xpl.where())
         =====      ANSYS File Xplorer : Display Current Location
         Current Location : FULL::MASS
            File Location : 7644
        """
        response = self._mapdl.run("*XPL,STEP,%s" % where)
        if 'Not Found' in response:
            raise RuntimeError(response.strip())
        return response

    def info(self, recname, option=""):
        """Gives details on a specific record, or all records (using
        ``"*"``)

        Parameters
        ----------
        recname : str
            Record of interest

        option : str
            Options string.

        Returns
        -------
        mapdl_response : str
            Response from MAPDL.

        Examples
        --------
        >>> xpl.open('file.full')
        >>> print(xpl.info('NGPH'))
        =====      ANSYS File Xplorer : Information about Block NGPH
        ::NGPH                 Size =      6.289 KB
                 - Record Size   : 81
                 - Data type     : integer values
        """
        return self._mapdl.run("*XPL,INFO," + recname + "," + option)

    def print(self, recname):
        """Print values of a given records, or all records (using
        ``"*"``)

        Parameters
        ----------
        recname : str
            Record of interest

        option : str
            Options string.

        Returns
        -------
        mapdl_response : str
            Response from MAPDL.

        Examples
        --------
        >>> xpl.open('file.full')
        >>> print(xpl.print('DOFSBYNOD'))
        =====      ANSYS File Xplorer : Print Block DOFSBYNOD
        DOFSBYNOD :
        Size : 3
               1         2         3

        """
        return self._mapdl.run("*XPL,PRINT,%s" % recname)

    def json(self):
        """Create a JSON representation of the tree or records.

        Examples
        --------
        >>> xpl.json()
        {'name': 'FULL',
         'children': [{'name': 'DOFSBYNOD', 'size': 24},
         {'name': 'BACK', 'size': 336},
         {'name': 'STIFF', 'size': 120132},
         {'name': 'RHS', 'size': 1956},
         {'name': 'DIAGK', 'size': 1956},
         {'name': 'SCLK', 'size': 36},
         {'name': 'NODEEXT', 'size': 32},
         {'name': 'PCGDOFS', 'size': 984},
         {'name': 'BCDOFS', 'size': 984},
         {'name': 'BCVALUES', 'size': 20},
         {'name': 'MASS', 'size': 52020},
         {'name': 'DIAGM', 'size': 1236},
         {'name': 'NGPH', 'size': 6440}]}
        """
        self._mapdl.run("*XPL,JSON,_mylocal_.json")
        text = self._mapdl._download_as_raw('_mylocal_.json').decode()
        return json.loads(text)

    def where(self):
        """Prints the current location in the MAPDL FIle

        Returns
        -------
        location_string : str
            String containing the current location.

        Examples
        --------
        >>> print(xpl.where())
         =====      ANSYS File Xplorer : Display Current Location
         Current Location : FULL
            File Location : 412
        """
        return self._mapdl.run("*XPL,WHERE")

    def up(self, nlev=1):
        """Go up in the tree.

        nlev : int
            Number of levels to recursively go up, or TOP

        Examples
        --------
        >>> print(xpl.up())
         =====      ANSYS File Xplorer : Go up to 1 level(s)
                     -> Already at the top level. Command is ignored
        """
        if str(nlev).upper().strip() == 'TOP':
            return self._mapdl.run("*XPL,UP,TOP")
        return self._mapdl.run("*XPL,UP,%d" % nlev)

    def goto(self, path):
        """Go directly to a new location in the file.

        Parameters
        ----------
        path : str
            Absolute path to the new location.

        Examples
        --------
        >>> print(xpl.goto('MASS'))
         =====      ANSYS File Xplorer : Go up to top level(s)
         =====      ANSYS File Xplorer : Step into Block MASS
        """
        return self._mapdl.run("*XPL,GOTO,%s" % path)

    def copy(self, newfile, option=''):
        """Copy the current opened as a new file.

        Parameters
        ----------
        newfile : str
            Name of the new file to create

        option: str
            Option.

        Examples
        --------
        >>> xpl.copy('tmpfile.full')
         =====      ANSYS File Xplorer : Copy file.full ANSYS file to file tmpfile.full
            >>      Remove existing output file tmpfile.full
        """
        return self._mapdl.run("*XPL,COPY,%s,%s" % (newfile, option))

    def save(self):
        """Save the current file, ignoring the marked records"""
        response = self._mapdl.run("*XPL,SAVE").strip()
        self._check_ignored(response)
        return response

    def read(self, recordname):
        """Read a given record and fill a Python array.

        Returns
        -------
        arr : ansys.mapdl.AnsMat
            The array of values.

        Examples
        --------
        >>> vec = xpl.read('MASS')
        >>> vec.asarray()
        array([ 4,  7, 10, 13, 16, 19, 22, 25, 28, 31, 34, 37, 40, 43,
               46, 49, 52, 55, 58,  1], dtype=int32)
        """
        response = self._mapdl.run("*XPL,READ,%s,TmpXplData" % recordname)
        self._check_ignored(response)
        response = self._mapdl._data_info("TmpXplData")

        if (response.stype == anskernel.INTEGER):
            dtype = np.int32
        elif (response.stype == anskernel.DOUBLE):
            dtype = np.double
        elif (response.stype == anskernel.HYPER):
            dtype = np.int64
        else:
            raise TypeError('Unhandled ANSYS type %s' % response.stype)

        mm = self._mapdl.math
        return mm.vec(dtype=dtype, name="TmpXplData")

    def __repr__(self):
        txt = 'MAPDL File Explorer\n'
        if self._open:
            txt += '\tOpen file:%s' % self._filename
            txt += '\n'.join(self.where().splitlines()[1:])
        else:
            txt += '\tNo open file'
        return txt



import appdirs
import os

_LOCAL_PORTS = []

# Per contract with Sphinx-Gallery, this method must be available at top level
from pyvista.utilities.sphinx_gallery import _get_sg_image_scraper

from ansys.mapdl.core._version import __version__
from ansys.mapdl.core.convert import convert_script
from ansys.mapdl.core.launcher import (launch_mapdl, change_default_ansys_path,
                                       close_all_local_instances)
from ansys.mapdl.core.misc import Report, _configure_pyvista, _check_has_ansys
from ansys.mapdl.core.launcher import get_ansys_path
from ansys.mapdl.core.mapdl_grpc import MapdlGrpc as Mapdl
from ansys.mapdl.core.pool import LocalMapdlPool

from ansys.mapdl.core import examples

# from ansys.mapdl.core.xpl import ansXpl
# from ansys.mapdl.core.math import MapdlMath

# from ansys.mapdl.core.mapdl_azure import MapdlAzure, JupyterAzureClient

_HAS_ANSYS = _check_has_ansys()

# Setup data directory
try:
    USER_DATA_PATH = appdirs.user_data_dir('ansys_mapdl_core')
    if not os.path.exists(USER_DATA_PATH):  # pragma: no cover
        os.makedirs(USER_DATA_PATH)

    EXAMPLES_PATH = os.path.join(USER_DATA_PATH, 'examples')
    if not os.path.exists(EXAMPLES_PATH):  # pragma: no cover
        os.makedirs(EXAMPLES_PATH)

except:  # pragma: no cover
    pass

# set pyvista defaults
_configure_pyvista()


if 'ANSJUPHUB_VER' in os.environ:
    from ansys.mapdl.core.jupyter import launch_mapdl_on_cluster as launch_mapdl



"""Module to support MAPDL CAD geometry"""
import re

from pyiges import Iges
import numpy as np
import pyvista as pv
from vtk import vtkAppendPolyData

from ansys.mapdl.core.misc import supress_logging, run_as_prep7


class Geometry():
    """Pythonic representation of MAPDL CAD geometry"""

    def __init__(self, mapdl):
        from ansys.mapdl.core.mapdl import _MapdlCore
        if not isinstance(mapdl, _MapdlCore):
            raise TypeError('Must be initialized using a MAPDL class')

        self._mapdl = mapdl
        self._keypoints_cache = None
        self._lines_cache = None
        self._log = self._mapdl._log

    def _set_log_level(self, level):
        return self._mapdl.set_log_level(level)

    def _load_iges(self):
        """Loads the iges file from MAPDL as a pyiges class"""
        return Iges(self._mapdl._generate_iges())

    def _reset_cache(self):
        self._keypoints_cache = None
        self._lines_cache = None

    @property
    def _keypoints(self):
        """Returns keypoints cache"""
        if self._keypoints_cache is None:
            self._keypoints_cache = self._load_keypoints()
        return self._keypoints_cache

    @property
    def keypoints(self):
        """Keypoint coordinates"""
        return np.asarray(self._keypoints.points)

    @property
    def _lines(self):
        """Returns lines cache"""
        if self._lines_cache is None:
            self._lines_cache = self._load_lines()
        return self._lines_cache

    @property
    def lines(self):
        """Active lines as a pyvista.PolyData"""
        return self._lines


    def areas(self, quality=7):
        """List of areas from MAPDL represented as ``pyvista.PolyData``.

        Parameters
        ----------
        quality : int, optional
            quality of the mesh to display.  Varies between 1 (worst)
            to 10 (best).

        Returns
        -------
        areas : list
            List of ``pyvista.PolyData`` areas representing the active
            surface areas selected by ``ASEL``.

        """
        surf = self.generate_surface(11 - quality)

        areas = []
        anums = np.unique(surf['entity_num'])
        for anum in anums:
            areas.append(surf.extract_cells(surf['entity_num'] == anum))
        return areas

    @supress_logging
    @run_as_prep7
    def generate_surface(self, density=5, amin=None, amax=None, ninc=None):
        """Generate an-triangular surface of the active surfaces.

        Parameters
        ----------
        density : int, optional
            APDL smart sizing option.  Ranges from 1 (worst) to 10
            (best).

        amin : int, optional
            Minimum APDL numbered area to select.  See
            ``mapdl.anum`` for available areas.

        amax : int, optional
            Maximum APDL numbered area to select.  See
            ``mapdl.anum`` for available areas.

        ninc : int, optional
            Steps to between amin and amax.
        """
        # store initially selected areas and elements
        self._mapdl.cm('__tmp_area__', 'AREA')
        self._mapdl.cm('__tmp_elem__', 'ELEM')

        # reselect from existing selection to mimic APDL behavior
        if amin or amax:
            if amax is None:
                amax = amin
            else:
                amax = ''

            if amin is None:  # amax is non-zero
                amin = 1

            if ninc is None:
                ninc = ''

            self._mapdl.asel('R', 'AREA', vmin=amin, vmax=amax, vinc=ninc)

        # duplicate areas to avoid affecting existing areas
        a_num = int(self._mapdl.get(entity='AREA', item1='NUM', it1num='MAXD'))
        with self._mapdl.chain_commands:
            self._mapdl.numstr('AREA', a_num)
            self._mapdl.agen(2, 'ALL', noelem=1)
        a_max = int(self._mapdl.get(entity='AREA', item1='NUM', it1num='MAXD'))

        with self._mapdl.chain_commands:
            self._mapdl.asel('S', 'AREA', vmin=a_num + 1, vmax=a_max)
            self._mapdl.aatt()  # necessary to reset element/area meshing association

        # create a temporary etype
        etype_max = int(self._mapdl.get(entity='ETYP', item1='NUM', it1num='MAX'))
        etype_old = self._mapdl.parameters.type
        etype_tmp = etype_max + 1

        with self._mapdl.chain_commands:
            self._mapdl.et(etype_tmp, 'MESH200', 4)
            self._mapdl.shpp('off')
            self._mapdl.smrtsize(density)
            self._mapdl.type(etype_tmp)

        if self._mapdl.parameters.routine != 'PREP7':
            self._mapdl.prep7()

        resp = self._mapdl.amesh('all')

        # must know the number of elements in each area
        reg = re.compile(r'Meshing of area (\d*) completed \*\* (\d*) elements')
        groups = reg.findall(resp)
        if not groups:
            reg = re.compile(r'AREA\s*(\d*).*?,\s*(\d*) TRIANGLES')
            groups = reg.findall(resp)
        groups = [[int(anum), int(nelem)] for anum, nelem in groups]
        self._mapdl.esla('S')

        grid = self._mapdl.mesh._grid.linear_copy()
        pd = pv.PolyData(grid.points, grid.cells)

        pd['ansys_node_num'] = grid['ansys_node_num']
        pd['vtkOriginalPointIds'] = grid['vtkOriginalPointIds']
        # pd.clean(inplace=True)  # OPTIONAL

        # delete all temporary meshes and clean up settings
        with self._mapdl.chain_commands:
            self._mapdl.aclear('ALL')
            self._mapdl.adele('ALL', kswp=1)
            self._mapdl.numstr('AREA', 1)
            self._mapdl.type(etype_old)
            self._mapdl.etdele(etype_tmp)
            self._mapdl.shpp('ON')
            self._mapdl.smrtsize('OFF')
            self._mapdl.cmsel('S', '__tmp_area__', 'AREA')
            self._mapdl.cmsel('S', '__tmp_elem__', 'ELEM')

        entity_num = np.empty(grid.n_cells, dtype=np.int32)
        if grid and groups:
            # add anum info
            i = 0
            for anum, nelem in groups:
                entity_num[i:i+nelem] = anum
                i += nelem
        else:
            entity_num[:] = 0

        pd['entity_num'] = entity_num
        return pd

    @property
    def n_volu(self):
        """Number of volumes currently selected

        Examples
        --------
        >>> mapdl.n_area
        1
        """
        return self._item_count('VOLU')

    @property
    def n_area(self):
        """Number of areas currently selected

        Examples
        --------
        >>> mapdl.n_area
        1
        """
        return self._item_count('AREA')

    @property
    def n_line(self):
        """Number of lines currently selected

        Examples
        --------
        >>> mapdl.n_line
        1
        """
        return self._item_count('LINE')

    @property
    def n_keypoint(self):
        """Number of keypoints currently selected

        Examples
        --------
        >>> mapdl.n_keypoint
        1
        """
        return self._item_count('KP')

    @supress_logging
    def _item_count(self, entity):
        """Return item count for a given entity"""
        return int(self._mapdl.get(entity=entity, item1='COUNT'))

    @property
    def knum(self):
        """Array of keypoint numbers.

        Examples
        --------
        >>> mapdl.block(0, 1, 0, 1, 0, 1)
        >>> mapdl.knum
        array([1, 2, 3, 4, 5, 6, 7, 8], dtype=int32)
        """
        return self._mapdl.get_array('KP', item1='KLIST').astype(np.int32)    

    @property
    def lnum(self):
        """Array of line numbers.

        Examples
        --------
        >>> mapdl.block(0, 1, 0, 1, 0, 1)
        >>> mapdl.lnum
        array([ 1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12], dtype=int32)
        """
        # this (weirdly) sometimes fails
        for _ in range(5):
            lnum = self._mapdl.get_array('LINES', item1='LLIST')
            if lnum.size == self.n_line:
                break
        return lnum.astype(np.int32)

    @property
    def anum(self):
        """Array of area numbers.
        Examples
        --------
        >>> mapdl.block(0, 1, 0, 1, 0, 1)
        >>> mapdl.anum
        array([1, 2, 3, 4, 5, 6], dtype=int32)
        """
        return self._mapdl.get_array('AREA', item1='ALIST').astype(np.int32)

    @property
    def vnum(self):
        """Array of volume numbers.

        Examples
        --------
        >>> mapdl.block(0, 1, 0, 1, 0, 1)
        >>> mapdl.vnum
        array([1], dtype=int32)
        """
        return self._mapdl.get_array('VOLU', item1='VLIST').astype(np.int32)

    @supress_logging
    def _load_lines(self):
        """Load lines from MAPDL using IGES"""
        # ignore volumes
        with self._mapdl.chain_commands:
            self._mapdl.cm('__tmp_volu__', 'VOLU')
            self._mapdl.cm('__tmp_area__', 'AREA')
            self._mapdl.cm('__tmp_keyp__', 'KP')
            self._mapdl.asel('S', 'ALL')
            self._mapdl.ksel('S', 'ALL')
            self._mapdl.vsel('NONE')

        iges = self._load_iges()

        with self._mapdl.chain_commands:
            self._mapdl.cmsel('S', '__tmp_volu__', 'VOLU')
            self._mapdl.cmsel('S', '__tmp_area__', 'AREA')
            self._mapdl.cmsel('S', '__tmp_keyp__', 'KP')

        selected_lnum = self.lnum
        lines = []
        entity_nums = []
        for bspline in iges.bsplines():
            # allow only 10001 as others appear to be construction entities
            if bspline.d['status_number'] == 10001:
                entity_num = int(bspline.d['entity_subs_num'])
                if entity_num not in entity_nums and entity_num in selected_lnum:
                    entity_nums.append(entity_num)
                    line = bspline.to_vtk()
                    line.cell_arrays['entity_num'] = entity_num
                    lines.append(line)

        entities = iges.lines() + iges.circular_arcs()
        for line in entities:
            if line.d['status_number'] == 1:
                entity_num = int(line.d['entity_subs_num'])
                if entity_num not in entity_nums and entity_num in selected_lnum:
                    entity_nums.append(entity_num)
                    line = line.to_vtk(resolution=100)
                    line.cell_arrays['entity_num'] = entity_num
                    lines.append(line)

        if lines:
            afilter = vtkAppendPolyData()
            for line in lines:
                afilter.AddInputData(line)
            afilter.Update()
            lines = pv.wrap(afilter.GetOutput())
            lines['entity_num'] = lines['entity_num'].astype(np.int32)
        else:
            lines = pv.PolyData()

        return lines

    def _load_keypoints(self):
        """Load keypoints from MAPDL using IGES"""
        # write only keypoints
        with self._mapdl.chain_commands:
            self._mapdl.cm('__tmp_volu__', 'VOLU')
            self._mapdl.cm('__tmp_area__', 'AREA')
            self._mapdl.cm('__tmp_line__', 'LINE')
            self._mapdl.vsel('NONE')
            self._mapdl.asel('NONE')
            self._mapdl.lsel('NONE')

        iges = self._load_iges()

        with self._mapdl.chain_commands:
            self._mapdl.cmsel('S', '__tmp_volu__', 'VOLU')
            self._mapdl.cmsel('S', '__tmp_area__', 'AREA')
            self._mapdl.cmsel('S', '__tmp_line__', 'LINE')

        keypoints = []
        kp_num = []
        for kp in iges.points():
            keypoints.append([kp.x, kp.y, kp.z])
            kp_num.append(int(kp.d['entity_subs_num']))

        # self._kp_num = np.array(self._kp_num)
        keypoints_pd = pv.PolyData(keypoints)
        keypoints_pd['entity_num'] = kp_num
        return keypoints_pd

    def __str__(self):
        """Current geometry info"""
        info = 'MAPDL Selected Geometry\n'
        info += 'Keypoints:  %d\n' % self.n_keypoint
        info += 'Lines:      %d\n' % self.n_line
        info += 'Areas:      %d\n' % self.n_area
        info += 'Volumes:    %d\n' % self.n_volu
        return info



"""Contains methods used only when running on ANSYS's jupyterhub cluster"""
import warnings

try:
    from ansys.jupyterhub import manager
except ImportError:
    raise ImportError('Module `ansys-jupyterhub-manager` missing.\n'
                      'This module is required to spawn instances on jupyterhub')


MAX_CPU = 128
MAX_MEM = 256


def check_manager():
    try:
        # response = manager.ping()
        manager.ping()
    except:
        raise RuntimeError('Unable to connect to scheduler')

    # consider checking the version
    # version = re.findall('(\d*\.\d*\.\d*)', response)
    # if not version:
    #     raise RuntimeError('Unable to parse version')
    # if version[0] != '0.1.6':
    #     raise RuntimeError('Invalid scheduler version')


def launch_mapdl_on_cluster(nproc=2, memory=4, loglevel='INFO',
                            additional_switches='', verbose=True,
                            start_timeout=600, **kwargs):
    """Start MAPDL on the ANSYS jupyter cluster in gRPC mode.

    Parameters
    ----------
    nproc : int, optional
        Number of processors.  Defaults to 2.

    memory : float, optional
        Fixed amount of memory to request for MAPDL in Gigabytes.  If
        the mapdl instance requires more ram than your provide MAPDL
        may segfault.

    loglevel : str, optional
        Sets which messages are printed to the console.  Default
        'INFO' prints out all ANSYS messages, 'WARNING` prints only
        messages containing ANSYS warnings, and 'ERROR' prints only
        error messages.

    additional_switches : str, optional
        Additional switches for MAPDL, for example aa_r, and academic
        research license, would be added with:

        - ``additional_switches="-aa_r"``

        Avoid adding switches like -i -o or -b as these are already
        included to start up the MAPDL server.  See the notes
        section for additional details.

    start_timeout : float, optional
        Maximum allowable time to connect to the MAPDL server.

    Returns
    -------
    port : int
        Returns the port number that the gRPC instance started on.

    Examples
    --------
    Launch MAPDL using the default configuration.

    >>> from ansys.mapdl import launch_mapdl
    >>> mapdl = launch_mapdl()
    """
    # attempt to connect to the remote scheduler
    check_manager()

    # check additional_switches args
    if '-m ' in additional_switches:
        raise ValueError('Memory option "-m" not permitted when launching from the '
                         'kubernetes cluster and is set with the `memory` parameter')
    if '-np ' in additional_switches:
        raise ValueError('CPU option "-np" not permitted when launching from the '
                         'kubernetes cluster and is set with the `nproc` parameter')

    # check resources
    nproc = int(nproc)
    if nproc < 0:
        raise ValueError('Requested CPUs `nproc` must be greater than 0')
    if nproc > MAX_CPU:
        raise ValueError(f'Requested CPUs `nproc` must be less than {MAX_CPU}')

    if memory < 0.25:
        raise ValueError('Requested memory `mem` must be greater than 0.25')
    if memory > MAX_MEM:
        raise ValueError(f'Requested memory `mem` must be less than than {MAX_MEM}')

    # convert memory from GB to Mi
    memory *= 1024
    if '-smp' in additional_switches:
        warnings.warn('Ignoring additional switch "-smp".  Incompatible with docker '
                      'container.')
        additional_switches = additional_switches.replace('-smp', '')
    additional_switches += f'-m -{memory} -np {nproc}'

    # need a way of making the image user-selectable
    image = 'mapdlhelm.azurecr.io/mapdl:v22.0.0'
    command = 'printf "" | /ansys_inc/v202/ansys/bin/mapdl %s -smp -grpc -custom /ansys_inc/v202/grpc/ansys.e201t.DEBUG-0.53.1' % additional_switches
    env = {'ANSYSLMD_LICENSE_FILE': '1055@10.0.0.16'}
    ip, pod_name = manager.spawn_pod(image, env=env, cpu=1000*nproc, memory=memory,
                                     command=command, start_timeout=start_timeout,
                                     verbose=verbose)

    # connect to the pod instance
    from ansys.mapdl import Mapdl  # import here to avoid recursive
    return Mapdl(ip, loglevel=loglevel)
